from setuptools import setup, find_packages


def get_description():
    return "Tool for populating ubi repositories"


def get_long_description():
    with open("README.md") as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find("\n\n")
    return text[idx:]


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="ubi-population-tool",
    version="0.14.0",
    license="GNU General Public License",
    author="",
    author_email="",
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    url="https://github.com/release-engineering/ubi-population-tool",
    install_requires=get_requirements(),
    python_requires=">=3.7",
    packages=find_packages(exclude=["tests", "tests.*"]),
    entry_points={
        "console_scripts": [
            "ubipop = ubipop.cli:entry_point",
        ]
    },
)
